# -*- coding: utf-8 -*-
"""GiocaOnline client module"""

import pytz
import logging

from datetime import datetime
from typing import Optional, Iterable, Mapping, List

from requests import Response

from giocaonline.client.base_aams_client_mixin import AAMSClientMixin
from giocaonline.constants import PrizePoolType
from giocaonline.exceptions import InternalError, AAMSException, ValidationError

logger = logging.getLogger(__name__)

GIOCAONLINE_DATETIME_FORMAT = "%d%m%Y%H%M%S"

PrizeObject = Mapping[str, float]
WinObject = Mapping[str, float]


class GiocaOnlineInternalError(InternalError):
    """Indica un errore di GiocaOnline rispetto alle specifiche indicate in documentazione."""


def strftime(date: datetime):
    return date.strftime(GIOCAONLINE_DATETIME_FORMAT)


class GiocaOnlineClient(AAMSClientMixin):
    """
    GiocaOnline HTTP client.

    Comunicazioni di gioco
    =========================

    Il modello qui utilizzato è quello che il PGDA deﬁnisce "modalità 2". Le iscrizioni vengono raccolte dopo
     la creazione del torneo (cioè l'apertura sessione).

    Ciascuna dà origine a una richiesta DIRITTO e può venire annullata con una richiesta ANNULLO DIRITTO.
    Al termine delle iscrizioni viene richiesta la CONVALIDA SESSIONE.

    La sequenza delle operazioni è quindi (in maiuscolo l'invocazione dei metodi della presente API):
    • INIZIO SESSIONE
    • inizio iscrizioni
    • DIRITTO / ANNULLO DIRITTO ...
    • ﬁne iscrizioni
    • CONVALIDA SESSIONE
    • PIANO PREMI
    • inizio gioco
    • ﬁne gioco
    • LISTA VINCITORI
    • ACCREDITO ...
    • FINE SESSIONE

    Modalità di trasmissione
    ========================

    Le API verranno invocate dal provider con chiamate HTTPS su un endpoint GOL prestabilito
     (<dominio>/servizi/API.php).
    Si assume che le richieste partano da sistemi identiﬁcati (tramite IP o subnet) e abbiano sempre un set di
    parametri obbligatori volti ad autorizzare e identiﬁcare il richiedente (deﬁniti separatamente da queste speciﬁche).
    Richiesta e risposta per ciascun metodo sono codiﬁcati come oggetti JSON. Nei nomi delle chiavi, le prime
    due lettere (minuscole) indicano la tipologia di dato, secondo la tabella:

    preﬁsso   -      descrizione
    -----------------------------------------------------------------------------------------------------------
    fl         -     valore booleano. Si utilizzano le costanti letterali "SI" e "NO", maiuscole
    id        -      identiﬁcativo alfanumerico composto esattamente da 16 caratteri
    ip        -      indirizzo IP in notazione dotted-quad
    is        -      identiﬁcativo alfanumerico composto esattamente da 14 caratteri
    ls        -      stringa da elenco
    ll        -      importi o valori numerici a 64 bit
    nm        -      valore numerico
    pc        -      valore percentuale, in centesimi (es. 5% => 500)
    sd        -      formato data ridotto. Il valore deve essere formattato come ddmmyyyy (giorno, mese, anno)
    st        -      stringa
    vl        -      valori rappresentanti importi, espressi in centesimi di euro
    xd        -      formato data esteso. Il valore deve essere formattato come ddmmyyyyhhmmss
                    (giorno, mese, anno, ore, minuti, secondi)

    """
    current_message: Optional[int] = None
    data: Optional[dict] = None

    # • richiesta saldo giocatore
    # • creazione sessione
    # • iscrizione a torneo
    # • cancellazione da torneo
    # • convalida sessione
    # • piano premi
    # • lista vincitori
    # • accredito vincita
    # • chiusura sessione

    _methods_map: dict = {
        # i metodi che iniziano con 1 non sono previsti da AAMS, ma soltanto da GiocaOnline per gestire le
        # comunicazioni fra provider del gioco e concessionario. La numerazione è completamente arbitraria
        # ed interna all' pacchetto python.

        100: 'autenticazione_giocatore_100',
        101: 'richiesta_saldo_giocatore_101',



        # Metodi AAMS
        200: 'inizio_sessione_200',
        220: 'acquisto_diritto_di_partecipazione_220',
        230: 'richiesta_annullamento_diritto_di_partecipazione_230',
        240: 'piano_dei_premi_240',
        250: 'convalida_della_sessione_250',
        260: 'lista_vincitori_260',
        280: 'accredito_vincita_280',
        #
        300: 'fine_sessione_300',
        # 310: 'richiesta_invalidazione_sessione_310',
        # 320: 'richiesta_invalidazione_diritto_partecipazione_320',
        # 330: 'verifica_stato_richiesta_invalidazione_330',
        # 340: 'accredito_rimborso_340',
        # 360: 'verifica_correttezza_sessioni_effettuate_360',
        # 365: 'richiesta_notifica_anomalie_riscontrate_per_sessione_365',
        #
        # 800: 'richiesta_dati_rendicondazione_contabile_800',
        # 810: 'richiesta_aggiornamento_data_fine_sessione_810',
        # 830: 'comunicazione_del_software_installato_830',
        # 831: 'comunicazione_del_software_installato_per_versione_831',
        # 840: 'aggiunta_del_certificato_di_firma_840',
    }

    endpoints: dict = {
        100: '',
        101: '',
        200: '',
        220: '',
        230: '',
        240: '',
        250: '',
        260: '',
        280: '',

        300: '',
        310: '',
        320: '',
        330: '',
        340: '',
        360: '',
        365: '',

        800: '',
        810: '',
        830: '',
        831: '',
        840: '',
    }

    # MESSAGGI INTERNI PROVIDER - GIOCAONLINE - CONCESSIONARIO
    def autenticazione_giocatore_100(
        self,
        player_token: str,
        player_account_code: str,
        dealer_code: int,
    ):
        """
        Richiesta di veriﬁca del token passato dal concessionario

        :param player_token: (stGiocatore) codice univoco (token) assegnato dal concessionario per il lancio del gioco
        :param player_account_code: (stConto) codice conto del giocatore
        :param dealer_code: (nmNetwork) codice numerico identiﬁcativo del concessionario del giocatore (deﬁnito da GOL)
        """
        self._do_request_message(
            message_code=100,
            player_token=player_token,
            player_account_code=player_account_code,
            dealer_code=dealer_code,
        )

    def validate_and_prepare_data_100(
        self,
        player_token: str,
        player_account_code: str,
        dealer_code: int,
    ):
        # TODO: validate dealer_code in Enum ?

        # PREPARE DATA
        self.data = {
            "stGiocatore": player_token,
            "stConto": player_account_code,
            "nmNetwork": dealer_code,
        }

    def handle_message_100_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 100.

        Prametri in risposta:
        =====================

        {
            "account_code": "abc", # (stConto) codice conto del giocatore
            "player_ssn": "abc", # (idCodFisc) codice ﬁscale del giocatore (o suo hash)
            "region": "abc", # (lsRegione) codice regione del giocatore
            "reseller_id": "abc", # (nmRivenditore) (opzionale) ID del rivenditore
            "access_token": "abc", # (stAccesso) codice di accesso (token) da impiegare per le successive richieste sulla piattaforma GOL,
        }

        NB: reseller_id (nmRivenditore) è un codice numerico (intero a 32 bit) utilizzabile dagli aggregatori
        per indicare il concessionario di appartenenza del giocatore.
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {
                    'account_code': msg_or_exception['stConto'],
                    'player_ssn': msg_or_exception['idCodFisc'],
                    'region': msg_or_exception['lsRegione'],
                    'access_token': msg_or_exception['stAccesso'],
                    'reseller_id': msg_or_exception.get('nmRivenditore', None),
                }
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# ================================================================================
    def richiesta_saldo_giocatore_101(
        self,
        gol_player_token: str,
        player_account_code: str,
        dealer_code: int,
    ):
        """
        Richiesta di saldo del conto giocatore

        :param gol_player_token: (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code: (stConto) codice conto del giocatore
        :param dealer_code: (nmNetwork) codice numerico identiﬁcativo del concessionario del giocatore (deﬁnito da GOL)
        """
        self._do_request_message(
            message_code=100,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            dealer_code=dealer_code,
        )

    def validate_and_prepare_data_101(
        self,
        gol_player_token: str,
        player_account_code: str,
        dealer_code: int,
    ):
        # TODO: validate dealer_code in Enum ?

        # PREPARE DATA
        self.data = {
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "nmNetwork": dealer_code,
        }

    def handle_message_101_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 101.

        Parametri in risposta:
        =====================

        {
            "account_amount": "11.22", # (vlSaldo) saldo del conto
            "account_bonus_amount": "11.22", # (vlSaldoPlayBonus) (opzionale) saldo play bonus
        }

        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {
                    'account_amount': msg_or_exception['vlSaldo'],
                    'account_bonus_amount': msg_or_exception.get('vlSaldoPlayBonus', None),
                }
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# ================================================================================
    # MESSAGGI PER LO SVOLGIMENTO DEL GIOCO IN MODALITA’ 2. A partire da pagina 30 della documentazione ufficiale.

    # 200
    def inizio_sessione_200(
        self,
        session_id: str,
        play_code: int,
        session_datetime_start: datetime,
        session_datetime_end: datetime,
        imp_diritto_partecip: float,
        prize_pool_type: PrizePoolType,
        prize_pool_perc: float,  # rtp
        num_expected_players: int = 0,  # valorizzare sempre a 0 per ADK (mai un SIT and GO)
        communication_modality: bool = True,
        minimum_guaranteed_sum: float = 0.0,
        buy_back_tickets_allowed: bool = False,
        is_bonus: bool = False,
        restricted_winnings: bool = False,  # True solo per i tornei SATELLITE
        int_jackpot: bool = False,
        additional_jackpot: bool = False,
        cup_modality: bool = False
    ):
        """
        Creazione di un nuovo torneo.

        :param session_id: (idSessConc) identiﬁcativo univoco della sessione sui sistemi del fornitore
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_datetime_start: (xdInizio) data inizio sessione (UTC)
        :param session_datetime_end: (sdFine) data ﬁne presunta sessione (UTC)
        :param communication_modality: (flModCondizionata) Valorizzare sempre a SI
        :param imp_diritto_partecip: (vlImporto) importo diritto di partecipazione
        :param prize_pool_type: (lsTpMtPr) tipologia di montepremi PERCENTUALE, VARIABILE, SOMMA_MIN_GARANTITA
        :param prize_pool_perc: (pcMontepremi) percentuale montepremi (es 5% --> 0.05)
        :param minimum_guaranteed_sum: (vlSommaMinGarantita) somma minima garantita
        :param buy_back_tickets_allowed: (lsRiacqBigl) sessione in cui sono consentiti acquisti multipli di biglietti.
            Passare un valore booleano che verrà trasformato in (NESSUNO (False) o REBUY (True))
        :param num_expected_players: (nmGiocatoriPrevisti) numero giocatori previsti
        :param is_bonus: (flBonus) sessione con bonus. Valorizzare a SI.
        :param restricted_winnings: (flSessVincColl) sessione con vincite vincolate
        :param int_jackpot: (flJckInterno) sessione con jackpot interno al gioco
        :param additional_jackpot: (flJckAgg)  sessione con jackpot aggiuntivi
        :param cup_modality: (flCup) sessione di gioco in modalità CUP. Valorizzare sempre a NO.

        Note
        ===========
        • Se lsTpMtPr=VARIABILE, pcMontepremi indica la percentuale media del montepremi (> 8000).
        • vlSommaMinGarantita è obbligatorio se lsTpMtPr=SOMMA_MIN_GARANTITA.
            # TODO posso passare di default 0 nei casi di torneo non garantito?
        • I valori consentiti per lsRiacqBigl sono NESSUNO (False) o REBUY (True).
        • nmGiocatoriPrevisti deve essere valorizzato al numero di giocatori previsti se il torneo è sit and go,
            altrimenti a 0.
        • flSessVincColl deve essere valorizzato a SI per i tornei satellite.

        """
        return self._do_request_message(
            message_code=200,
            session_id=session_id,
            play_code=play_code,
            session_datetime_start=session_datetime_start,
            session_datetime_end=session_datetime_end,
            communication_modality=communication_modality,
            imp_diritto_partecip=imp_diritto_partecip,
            prize_pool_type=prize_pool_type,
            prize_pool_perc=prize_pool_perc,
            minimum_guaranteed_sum=minimum_guaranteed_sum,
            buy_back_tickets_allowed=buy_back_tickets_allowed,
            num_expected_players=num_expected_players,
            is_bonus=is_bonus,
            restricted_winnings=restricted_winnings,
            int_jackpot=int_jackpot,
            additional_jackpot=additional_jackpot,
            cup_modality=cup_modality,
        )

    def validate_and_prepare_data_200(
        self,
        session_id: str,
        play_code: int,
        session_datetime_start: datetime,
        session_datetime_end: datetime,
        imp_diritto_partecip: float,
        prize_pool_type: PrizePoolType,
        prize_pool_perc: float,  # rtp
        num_expected_players: int,
        communication_modality: bool = True,
        minimum_guaranteed_sum: float = 0.0,
        buy_back_tickets_allowed: bool = False,
        is_bonus: bool = False,
        restricted_winnings: bool = False,
        int_jackpot: bool = False,
        additional_jackpot: bool = False,
        cup_modality: bool = False
    ):
        # session_id deve essere composta da esattamente 16 caratteri
        self.check_length("session_id (idSessConc)", session_id, 16, mode="e")
        # controlla che i valori datetime siano in formato UTC
        self.check_timezone('session_datetime_start (xdInizio)', session_datetime_start, pytz.utc)
        self.check_timezone('session_datetime_end (xdInizio)', session_datetime_end, pytz.utc)
        # controlla che l'importo sia maggiore o uguale a 0 e che non abbia più di due decimali.
        self.check_numeric_value('imp_diritto_partecip (vlImporto)', imp_diritto_partecip, 0, "ge")
        self.check_decimal_digits('imp_diritto_partecip (vlImporto)', imp_diritto_partecip, 2, "le")
        # verifica che il parametro communication_modality sia esattamente uguale a "SI"
        self.check_value_in('communication_modality (flModCondizionata)', communication_modality, [True])
        # verifica che la tipologia di montepremi sia una delle opzioni valide
        self.check_enum('prize_pool_type (lsTpMtPr)', prize_pool_type, PrizePoolType)
        # verifica che, se il tipo di torneo è a somma minima garantita, allora il valore di
        # `minimum_guaranteed_sum` (vlSommaMinGarantita) deve essere maggiore di zero, altrimenti 0
        if prize_pool_type == PrizePoolType.MINIMUM_GUARANTEED_SUM:
            self.check_numeric_value('minimum_guaranteed_sum (vlSommaMinGarantita)', minimum_guaranteed_sum, 0, "g")
        else:
            self.check_numeric_value('minimum_guaranteed_sum (vlSommaMinGarantita)', minimum_guaranteed_sum, 0, "e")
        # verifica che, `minimum_guaranteed_sum` abbia al massimo 2 decimali
        self.check_decimal_digits('minimum_guaranteed_sum (vlSommaMinGarantita)', minimum_guaranteed_sum, 2, "le")
        # verifica che il numero di giocatori previsti sia superiore o uguale a 0
        self.check_numeric_value('num_expected_players (nmGiocatoriPrevisti)', num_expected_players, 0, "ge")
        # verifica che il parametro cup_modality sia esattamente uguale a "NO"
        self.check_value_in('cup_modality (flModCondizionata)', cup_modality, [False])
        # verifica e trasforma i valori booleani
        assert isinstance(buy_back_tickets_allowed, bool), "buy_back_tickets_allowed must be a boolean param"
        buy_back_tickets_allowed = self.convert_bool_to_fl(buy_back_tickets_allowed)
        communication_modality = self.convert_bool_to_fl(communication_modality)
        is_bonus = self.convert_bool_to_fl(is_bonus)
        restricted_winnings = self.convert_bool_to_fl(restricted_winnings)
        int_jackpot = self.convert_bool_to_fl(int_jackpot)
        additional_jackpot = self.convert_bool_to_fl(additional_jackpot)
        cup_modality = self.convert_bool_to_fl(cup_modality)

        # PREPARE DATA
        self.data = {
            "idSessConc": session_id,
            "nmCodiceGioco": play_code,
            "xdInizio": strftime(session_datetime_start),
            "sdFine": strftime(session_datetime_end),
            "flModCondizionata": communication_modality,
            "vlImporto": imp_diritto_partecip,
            "lsTpMtPr": prize_pool_type,
            "pcMontepremi": self.convert_float_perc_to_cent_perc(prize_pool_perc),
            "vlSommaMinGarantita": minimum_guaranteed_sum,
            "lsRiacqBigl": "REBUY" if buy_back_tickets_allowed else "NESSUNO",
            "nmGiocatoriPrevisti": num_expected_players,
            "flBonus": is_bonus,
            "flSessVincColl": restricted_winnings,
            "flJckInterno": int_jackpot,
            "flJckAgg": additional_jackpot,
            "flCup": cup_modality,
        }

    def handle_message_200_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 200.

        Prametri in risposta:
        =====================

        session_id (idSessAams) identiﬁcativo univoco della sessione sui sistemi AAMS
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {'session_id': msg_or_exception['idSessAams']}
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# =============================================================================================================
    # 220
    def acquisto_diritto_di_partecipazione_220(
        self,
        play_code: int,
        dealer_code: int,
        gol_player_token: str,
        player_account_code: str,
        session_id: str,
        region: str,
        price: float,
        rebuy: str,
        player_ip: str,
        player_nickname: str,
        purchase_datetime: datetime,
        related_session_id: str,
        amount_for_jackpot: float,
        rake: float,
        tournament_id: int,
        device: int,
        pay_with_token: bool = False,
    ):
        """
        Richiesta di acquisto di una schedina.

        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param dealer_code: (nmNetwork) codice numerico identiﬁcativo del concessionario del giocatore (deﬁnito da GOL)
        :param gol_player_token: (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code: (stConto) codice conto del giocatore
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param region: (lsRegione) codice regione del giocatore
        :param price: (vlImporto) importo diritto partecipazione
        :param rebuy: (lsRiacqBigl) riacquisto
        :param player_ip: (ipGiocatore) indirizzo IP del giocatore
        :param player_nickname: (stAlias) alias giocatore
        :param purchase_datetime: (xdAcquisto) data acquisto diritto (UTC)
        :param related_session_id: (idSessCollegata) codice identificativo AAMS della sessione collegata (torneo satellite vinto)
        :param amount_for_jackpot: (vlImpoJck) importo destinato al fondo jackpot
        :param token: (flToken) acquisto con token
        :param rake: (vlRake) importo rake
        :param tournament_id: (stTorneo) codice univoco del torneo sui sistemi del provider
        :param device: (nmTipoDisp) tipo dispositivo

        Note
        ===========
            • idSessCollegata è usato nel caso di iscrizione ad un torneo principale dopo la vittoria in un torneo satellite, e
              rappresenta il codice sessione di quest'ultimo
            • lsRiacqBigl può valere NESSUNO o, per i tornei che consentono il riacquisto, REBUY
            • nmTipoDisp contiene la descrizione del dispositivo utilizzato dall'utente per l'accesso al gioco, secondo la codifica
              dettagliata nella documentazione PGDA ("Tabella dei tipi di terminali di accettazione gioco")
        """
        return self._do_request_message(
            message_code=220,
            play_code=play_code,
            dealer_code=dealer_code,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            session_id=session_id,
            region=region,
            price=price,
            rebuy=rebuy,
            player_ip=player_ip,
            player_nickname=player_nickname,
            purchase_datetime=purchase_datetime,
            related_session_id=related_session_id,
            amount_for_jackpot=amount_for_jackpot,
            rake=rake,
            tournament_id=tournament_id,
            device=device,
            pay_with_token=pay_with_token
        )

    def validate_and_prepare_data_220(
        self,
        play_code: int,
        dealer_code: int,
        gol_player_token: str,
        player_account_code: str,
        session_id: str,
        region: str,
        price: float,
        rebuy: str,
        player_ip: str,
        player_nickname: str,
        purchase_datetime: datetime,
        related_session_id: str,
        amount_for_jackpot: float,
        rake: float,
        tournament_id: int,
        device: int,
        pay_with_token: bool = False,
    ):
        # controlla che i valori datetime siano in formato UTC
        self.check_timezone('purchase_datetime (xdAcquisto)', purchase_datetime, pytz.utc)
        # controlla che l'importo sia maggiore o uguale a 0 e che non abbia più di due decimali.
        self.check_numeric_value('price (vlImporto)', price, 0, "ge")

        # PREPARE DATA
        self.data = {
            "nmCodiceGioco": play_code,
            "nmNetwork": dealer_code,
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "idSessAams": session_id,
            "lsRegione": region,
            "vlImporto": price,
            "lsRiacqBigl": rebuy,
            "ipGiocatore": player_ip,
            "stAlias": player_nickname,
            "xdAcquisto": purchase_datetime,
            "idSessCollegata": related_session_id,
            "vlImpoJck": amount_for_jackpot,
            "vlRake": rake,
            "stTorneo": tournament_id,
            "nmTipoDisp": device,
            "flToken": pay_with_token,
        }

    def handle_message_220_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 220.

        Prametri in risposta:
        =====================

        session_id (idSessAams) identiﬁcativo univoco della sessione sui sistemi AAMS
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    'played_aams_id': msg_or_exception['idBiglietto'],
                    'account_amount': msg_or_exception['vlSaldoCarta'],
                    'received_by_aams_datetime': msg_or_exception['sdTimeStamp'],
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# =============================================================================================================
    # 230
    def richiesta_annullamento_diritto_di_partecipazione_230(
            self,
            play_code: int,
            dealer_code: int,
            gol_player_token: str,
            player_account_code: str,
            session_id: str,
            played_aams_id: str,
            token: bool = False,
    ):
        """
        Richiesta di annullamento di un biglietto acquistato in precedenza.


        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param dealer_code: (nmNetwork) codice numerico identiﬁcativo del concessionario del giocatore (deﬁnito da GOL)
        :param gol_player_token: (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code: (stConto) codice conto del giocatore
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param played_aams_id: (idBiglietto) identificativo diritto di partecipazione assegnato dal sistema AAMS
        :param token: (flToken) acquisto con token

        """
        return self._do_request_message(
            message_code=230,
            play_code=play_code,
            dealer_code=dealer_code,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            session_id=session_id,
            played_aams_id=played_aams_id,
            token=token
        )

    def validate_and_prepare_data_230(
            self,
            play_code: int,
            dealer_code: int,
            gol_player_token: str,
            player_account_code: str,
            session_id: str,
            played_aams_id: str,
            token: bool = False,
    ):
        # controlla che il session_id sia esattamente di 16 caratteri
        self.check_length('session_id (idSessAams)', session_id, 16, 'e')
        # controlla che il played_aams_id sia esattamente di 16 caratteri
        self.check_length('played_aams_id (idBiglietto)', played_aams_id, 16, 'e')

        # PREPARE DATA
        self.data = {
            "nmCodiceGioco": play_code,
            "nmNetwork": dealer_code,
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "idSessAams": session_id,
            "idBiglietto": played_aams_id,
            "flToken": token,
        }

    def handle_message_230_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 230.

        Prametri in risposta:
        =====================

        account_amount (vlSaldoCarta) saldo conto di gioco
        received_by_aams_datetime (sdTimeStamp) data di acquisizione da parte di AAMS
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    'account_amount': msg_or_exception['vlSaldoCarta'],
                    'received_by_aams_datetime': msg_or_exception['sdTimeStamp'],
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# =============================================================================================================
    # 240
    def piano_dei_premi_240(
            self,
            play_code: int,
            session_id: str,
            prize_pool_perc: float,
            prize_pool_tot: float,
            internal_jackpot_tot: float,
            total_add_jackpot: float,
            total_prize_number: int,
            initial_prize: int,
            final_prize: int,
            prize_pool_part: float,
            internal_jackpot_part: float,
            partial_add_jackpot: float,
            rectification_progression_number: int,
            prizes_list: Iterable[PrizeObject]
    ):
        """
        Definizione dei premi da distribuire al termine del gioco


        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param prize_pool_perc (pcMontepremi) percentuale montepremi erogato
        :param prize_pool_tot (vlMontepremiTot) importo totale del montepremi erogato, comprensivo di eventuali jackpot
        :param internal_jackpot_tot (vlImpoJckInterno) quota del montepremi totale derivante da jackpot interno al gioco
        :param total_add_jackpot (vlImpoJckAgg) quota del montepremi totale derivante da jackpot aggiuntivi
        :param total_prize_number(nmTotPremi) numero totale premi da attribuire nella sessione di gioco
        :param initial_prize (nmPremioIniz) primo premio relativo all'invio corrente
        :param final_prize (nmPremioFinale) ultimo premio relativo all'invio corrente
        :param prize_pool_part (vlMontepremiParz) importo montepremi relativo ai premi presenti nell'invio
        :param internal_jackpot_part (vlJckInternoParz) quota del montepremi parziale derivante da jackpot interno al gioco
        :param partial_add_jackpot (vlJckAggParz) quota del montepremi parziale derivante da jackpot aggiuntivi
        :param rectification_progression_number (nmProgrRettifica) 0 per il primo invio del piano premi, va incrementato di 1 per ogni rettifica
        :param prizes_list (premi) lista di oggetti premio (ciascuno rappresenta un premio)

        Struttura oggetto premio
        ========================
        prize_pool_value (vlImpoMontepremi) valore del premio, comprensivo di eventuali jackpot
        internal_jackpot_value (vlImpoJckInterno) quota del premio derivante da jackpot interno al gioco
        additional_jackpot_value (vlImpoJckAgg) quota del premio derivante da jackpot aggiuntivi

        Note
        ===========
            • I premi presenti nella lista premi vanno indicati in ordine di importo decrescente.
            • Nel caso i premi siano più di 1000, sarà necessario spezzare la lista su più invii, ciascuno con al più 1000 premi.

        """
        return self._do_request_message(
            message_code=240,
            play_code=play_code,
            session_id=session_id,
            prize_pool_perc=prize_pool_perc,
            prize_pool_tot=prize_pool_tot,
            internal_jackpot_tot=internal_jackpot_tot,
            total_add_jackpot=total_add_jackpot,
            total_prize_number=total_prize_number,
            initial_prize=initial_prize,
            final_prize=final_prize,
            prize_pool_part=prize_pool_part,
            internal_jackpot_part=internal_jackpot_part,
            partial_add_jackpot=partial_add_jackpot,
            rectification_progression_number=rectification_progression_number,
            prizes_list=prizes_list,
        )

    def validate_and_prepare_data_240(
            self,
            play_code: int,
            session_id: str,
            prize_pool_perc: float,
            prize_pool_tot: float,
            internal_jackpot_tot: float,
            total_add_jackpot: float,
            total_prize_number: int,
            initial_prize: int,
            final_prize: int,
            prize_pool_part: float,
            internal_jackpot_part: float,
            partial_add_jackpot: float,
            rectification_progression_number: int,
            prizes_list: Iterable[PrizeObject],
    ):
        # Controlla che la lista premi non contenga più di 1000 elementi
        self.check_prizes_list('prizes_list (premi)', prizes_list, 1000)
        # Ordina la lista premi in ordine decrescente rispetto al valore del montepremi
        prizes_list.sort(key=lambda prize: prize.get('prize_pool_value'), reverse=True)

        # PREPARE DATA
        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "pcMontepremi": prize_pool_perc,
            "vlMontepremiTot": prize_pool_tot,
            "vlImpoJckInterno": internal_jackpot_tot,
            "vlImpoJckAgg": total_add_jackpot,
            "nmTotPremi": total_prize_number,
            "nmPremioIniz": initial_prize,
            "nmPremioFinale": final_prize,
            "vlMontepremiParz": prize_pool_part,
            "vlJckInternoParz": internal_jackpot_part,
            "vlJckAggParz": partial_add_jackpot,
            "nmProgrRettifica": rectification_progression_number,
            "premi": prizes_list,
        }

    def handle_message_240_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 240.

        Prametri in risposta:
        =====================

        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {}
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# =============================================================================================================
    # 250
    def convalida_della_sessione_250(
            self,
            play_code: int,
            session_id: str,
            validation_date: datetime,
    ):
        """
        Inviato alla partenza effettiva del torneo.


        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param validation_date: (xdConvalida) timestamp della convalida sessione (UTC)

        """
        return self._do_request_message(
            message_code=250,
            play_code=play_code,
            session_id=session_id,
            validation_date=validation_date,
        )

    def validate_and_prepare_data_250(
            self,
            play_code: int,
            session_id: str,
            validation_date: datetime,
    ):
        # controlla che il session_id sia esattamente di 16 caratteri
        self.check_length('session_id (idSessAams)', session_id, 16, 'e')
        # controlla che i valori datetime siano in formato UTC
        self.check_timezone('validation_date (xdConvalida)', validation_date, pytz.utc)

        # PREPARE DATA
        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "xdConvalida": validation_date,
        }

    def handle_message_250_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 250.

        Prametri in risposta:
        =====================

        received_by_aams_datetime (sdTimeStamp) data di acquisizione da parte di AAMS
        number_of_participation_validated (nmPartConvalidati) numero di partecipazioni convalidate da AAMS
        total_amount_of_participation_validated (vlImpPartConvalidati) importo totale raccolto con le partecipazioni convalidate
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    'received_by_aams_datetime': msg_or_exception['sdTimeStamp'],
                    'number_of_participation_validated': msg_or_exception['nmPartConvalidati'],
                    'total_amount_of_participation_validated': msg_or_exception['vlImpPartConvalidati'],
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# =============================================================================================================
    # 260
    def lista_vincitori_260(
        self,
        play_code: int,
        session_id: str,
        total_winner_number: int,
        initial_winner: int,
        final_winner: int,
        partial_win_prize: float,
        partial_internal_jackpot: float,
        partial_additional_jackpot: float,
        rectification_progressive_number: int,
        total_win_prize: float,
        total_internal_jackpot: float,
        total_additiona_jackpot: float,
        win_list: List[WinObject],
    ):
        """
        Comunicazione delle vincite realizzate in una sessione di gioco


        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param total_winner_number (nmTotVincitori) numero totale vincitori nella sessione di gioco
        :param initial_winner (nmInizVincita) progressivo prima vincita relativa al presente invio
        :param final_winner (nmFinVincita) progressivo ultima vincita relativa al presente invio
        :param partial_win_prize (vlImpoParzVincite) importo relativo alle vincite, comprensivo di jackpot e bonus
        :param partial_internal_jackpot (vlImpoParzJckInterno) quota delle vincite derivanti da jackpot interno al gioco
        :param partial_additional_jackpot (vlImpoParzJckAgg) quota delle vincite derivanti da jackpot aggiuntivo
        :param rectification_progressive_number (nmProgrRettifica) 0 per il primo invio, va incrementato di 1 per ogni richiesta di rettifica
        :param total_win_prize (vlImpoTotVincite) importo totale delle vincite erogate comprensivo di jackpot e bonus
        :param total_internal_jackpot (vlImpoTotJckInterno) quota delle vincite erogate derivante da jackpot interno al gioco
        :param total_additiona_jackpot (vlImpoTotJckAgg) quota delle vincite erogate derivante da jackpot aggiuntivo
        :param win_list (vincite) lista di oggetti vincita

        Struttura oggetto vincita
        =========================
        total_prize (vlImpoTotale) importo totale della vincita compresi jackpot e bonus
        internal_jackpot (vlImpoJckInterno) quota della vincita derivante da jackpot interno al gioco
        additional_jackpot (vlImpoJckAgg) quota della vincita derivante da jackpot aggiuntivo
        played_aams_id (idBiglietto) codice biglietto

        Note
        ========
            • Le vincite presenti nella lista vincite vanno indicate in ordine di importo totale decrescente.
            • Nel caso le vincite siano più di 1000, sarà necessario spezzare la lista su più invii, ciascuno con al più 1000 vincite.
        """
        return self._do_request_message(
            message_code=260,
            play_code=play_code,
            session_id=session_id,
            total_winner_number=total_winner_number,
            initial_winner=initial_winner,
            final_winner=final_winner,
            partial_win_prize=partial_win_prize,
            partial_internal_jackpot=partial_internal_jackpot,
            partial_additional_jackpot=partial_additional_jackpot,
            rectification_progressive_number=rectification_progressive_number,
            total_win_prize=total_win_prize,
            total_internal_jackpot=total_internal_jackpot,
            total_additiona_jackpot=total_additiona_jackpot,
            win_list=win_list,
        )

    def validate_and_prepare_data_260(
        self,
        play_code: int,
        session_id: str,
        total_winner_number: int,
        initial_winner: int,
        final_winner: int,
        partial_win_prize: float,
        partial_internal_jackpot: float,
        partial_additional_jackpot: float,
        rectification_progressive_number: int,
        total_win_prize: float,
        total_internal_jackpot: float,
        total_additiona_jackpot: float,
        win_list: List[WinObject],
    ):
        # Controlla che la lista vincitori non contenga più di 1000 elementi
        self.check_prizes_list('win_list (vincite)', win_list, 1000)

        # PREPARE DATA
        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "nmTotVincitori": total_winner_number,
            "nmInizVincita": initial_winner,
            "nmFinVincita": final_winner,
            "vlImpoParzVincite": partial_win_prize,
            "vlImpoParzJckInterno": partial_internal_jackpot,
            "vlImpoParzJckAgg": partial_additional_jackpot,
            "nmProgrRettifica": rectification_progressive_number,
            "vlImpoTotVincite": total_win_prize,
            "vlImpoTotJckInterno": total_internal_jackpot,
            "vlImpoTotJckAgg": total_additiona_jackpot,
            "vincite": self.prepare_winner_list(win_list),
        }

    def handle_message_260_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 260.

        Prametri in risposta:
        =====================
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {}
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

    @staticmethod
    def prepare_winner_list(win_list: List[WinObject]) -> List:
        # Ordina la lista vincitori in ordine decrescente rispetto all'importo totale della vincita
        win_list.sort(key=lambda win: win.get('total_prize'), reverse=True)
        try:
            return [
                {
                    'idBiglietto': win['played_aams_id'],
                    'vlImpoTotale': win['total_prize'],
                    'vlImpoJckInterno': win.get('internal_jackpot', 0),
                    'vlImpoJckAgg': win.get('additional_jackpot', 0),
                }
                for win in win_list
            ]
        except KeyError as e:
            raise ValidationError({str(e): "This key is required"})

# =============================================================================================================
    # 280
    def accredito_vincita_280(
            self,
            play_code: int,
            session_id: str,
            gol_player_token: str,
            player_account_code: str,
            credit_value: float,
            credit_date: datetime,
            credit_progressive_number: int,
            played_aams_id: str,
    ):
        """
        Richiesta di pagamento della vincita tramite accredito sul conto di gioco

        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param gol_player_token (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code (stConto) codice conto del giocatore
        :param credit_value (vlImpoVincita) importo accreditato
        :param credit_date (xdAccredito) data accredito vincita (UTC)
        :param credit_progressive_number (nmProgrAccredito) progressivo di accredito. Vale 1 per il primo accredito
        :param played_aams_id (idBiglietto) codice univoco del diritto di partecipazione assegnato dal sistema AAMS

        """
        return self._do_request_message(
            message_code=280,
            play_code=play_code,
            session_id=session_id,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            credit_value=credit_value,
            credit_date=credit_date,
            credit_progressive_number=credit_progressive_number,
            played_aams_id=played_aams_id,
        )

    def validate_and_prepare_data_280(
            self,
            play_code: int,
            session_id: str,
            gol_player_token: str,
            player_account_code: str,
            credit_value: float,
            credit_date: datetime,
            credit_progressive_number: int,
            played_aams_id: str,
    ):
        # session_id deve essere composta da esattamente 16 caratteri
        self.check_length("session_id (idSessConc)", session_id, 16, mode="e")
        # controlla che i valori datetime siano in formato UTC
        self.check_timezone('credit_date (xdAccredito)', credit_date, pytz.utc)
        # played_aams_id deve essere composta da esattamente 16 caratteri
        self.check_length("played_aams_id (idBiglietto)", played_aams_id, 16, mode="e")

        # PREPARE DATA
        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "vlImpoVincita": credit_value,
            "xdAccredito": credit_date,
            "nmProgrAccredito": credit_progressive_number,
            "idBiglietto": played_aams_id,
        }

    def handle_message_280_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 280.

        Prametri in risposta:
        =====================
        account_amount (vlSaldoCarta) saldo conto di gioco
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    'account_amount': msg_or_exception['vlSaldoCarta'],
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# =============================================================================================================
    # 300
    def fine_sessione_300(
            self,
            play_code: int,
            session_id: str,
            close_session_datetime: datetime,
            was_session_valid: bool = False,
    ):
        """
        Conclude la sessione di gioco

        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param close_session_datetime (xdChiusura) timestamp della chiusura sessione (UTC)
        :param was_session_valid (flChiusuraSess) SI se per la sessione era stata inviata la convalida, NO altrimenti

        """
        return self._do_request_message(
            message_code=300,
            play_code=play_code,
            session_id=session_id,
            close_session_datetime=close_session_datetime,
            was_session_valid=was_session_valid,
        )

    def validate_and_prepare_data_300(
            self,
            play_code: int,
            session_id: str,
            close_session_datetime: datetime,
            was_session_valid: bool = False,
    ):

        # session_id deve essere composta da esattamente 16 caratteri
        self.check_length("session_id (idSessAams)", session_id, 16, mode="e")
        # controlla che i valori datetime siano in formato UTC
        self.check_timezone('close_session_datetime (xdChiusura)', close_session_datetime, pytz.utc)
        was_session_valid = self.convert_bool_to_fl(was_session_valid)

        # PREPARE DATA
        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "xdChiusura": close_session_datetime,
            "flChiusuraSess": was_session_valid,
        }

    def handle_message_300_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 300.

        Prametri in risposta:
        =====================
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {}
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)

        raise msg_or_exception

# ================================================================================================================

    # RESPONSE HANDLING
    def handle_generic_message_response(self, response: Response) -> tuple:
        """
        Ogni richiesta prevede una risposta.

        Le tipologie di risposta sono due:
        • esito positivo: la risposta contiene lsEsito="OK", ed eventualmente altre chiavi che ne costituiscono
            il payload;
        • esito negativo: la risposta contiene solo le chiavi lsEsito="ERROR", nmFaultCode e stFaultDescription;
        le ultime due contengono rispettivamente il codice numerico dell'errore e una descrizione testuale dell'errore.

        """
        result = self.decode_response_body(response)

        try:
            esito = result.pop('lsEsito')
        except KeyError:
            self.raise_response_missing_field_error("lsEsito", result)

        if esito == "OK":  # noqa
            return True, result

        elif esito == "ERROR":
            try:
                error_code = result.pop("nmFaultCode")
            except KeyError:
                self.raise_response_missing_field_error("nmFaultCode", result)

            try:
                error_description = result.pop("stFaultDescription")
            except KeyError:
                self.raise_response_missing_field_error("stFaultDescription", result)

            return False, self.get_aams_message_error(error_code, error_description)  # noqa
        else:
            exc = GiocaOnlineInternalError('Expected GiocaOnline response body to have required key `lsEsito`'
                                           'equal to one of the following choices: "OK" or "ERROR".'
                                           f'During processing of message {self.current_message} this key was equal to:'
                                           f' {esito}.\nComplete Response body: {result}')
            logger.error(exc)
            raise exc

    # GiocaOnline ci passa la descrizione dell'errore.
    # Questo implica che non è necessario memorizzare tutte le descrizioni/codici di errore in quanto li gestisce
    # GiocaOnline.
    @staticmethod
    def get_aams_message_error(error_code: int, error_description) -> AAMSException:  # noqa
        return AAMSException(error_code, error_description)

    def raise_response_missing_field_error(self, key, response_body):
        exc = GiocaOnlineInternalError(
            f'Expected GiocaOnline response body to have required key `{key}`'
            f'During processing of message {self.current_message} this key was missing.'
            f'Complete Response body: {response_body}'
        )
        logger.error(exc)
        raise exc

    def convert_float_perc_to_cent_perc(self, perc: float):
        """
        Converte la percentuale di tipo float in percentuale centesimale.
        Ad esempio: 5% --> 0.05 --> 500
        """
        self.check_decimal_digits("Any percentage value", perc, 4, "le")
        perc = perc * 10000
        return int(perc)

    @staticmethod
    def convert_bool_to_fl(value: bool):
        """
        Converte il valore booleano nelle stringhe "SI" o "NO"
        """
        assert isinstance(value, bool), f"{value} is not of boolean type"
        return "SI" if value else "NO"
